/* $LAAS: server.c 2011/05/18 18:02:20 mallet $ */

/*
 * Copyright (c) 2008-2009,2011 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *					Anthony Mallet on Thu Jul  3 2008
 */

#define WINVER WindowsXP
#include <w32api.h>

#include <conio.h>
#include <stdio.h>
#include <getopt.h>
#include <windows.h>
#include <winsock2.h>

#include "EVaRT2.h"
#include "evart-client.h"

#include "internal.h"


/* --- local data ---------------------------------------------------- */

#define EVAS_N_KEEPALIVE	(16384) /* number of messages before sending a
                                         * keepalive request. About 80s @ 200Hz
                                         */
#define EVAS_TH_KEEPALIVE	(2048)	/* low threshold trigger */
#define EVAS_F_KEEPALIVE	(256)	/* resend frequency */

static SOCKET servfd;

static struct {
  evas_msgtype_t type;
  struct sockaddr addr;
  uint32_t keepalive;
  char name[128];
} requests[64];

static int	evas_serverlisten(SOCKET sockfd);
static void	evas_datahandler(sFrameOfData* FrameOfData);
static int	evas_requestnew(evas_msgtype_t type, struct sockaddr *addr,
			const char *name);
static int	evas_requestfree(evas_msgtype_t type, struct sockaddr *addr,
			const char *name);
static int	evas_requestrefresh(struct sockaddr *addr);
static void	evas_gettimeofacq(float delay, uint32_t *tv_sec,
			uint32_t *tv_usec);
static const char *
		evas_addrtoa(const struct sockaddr *addr);


/* --- main ---------------------------------------------------------- */

int
main(int argc, char* argv[])
{
  WSADATA supported;
  int s;

  /* options */
  static const char *usage =
    "usage: %s [--host evart_host] [--localhost this_host]\n"
    "\t\t\t[--port port_number]\n";
  static const struct option long_options[] = {
    { "host", 1, NULL, 'h' },
    { "localhost", 1, NULL, 'l' },
    { "port", 1, NULL, 'p' },
    { NULL, 0, NULL, 0 }
  };
  char *host = NULL;
  char *localhost = NULL;
  unsigned short port = 3615;

  while(1) {
    s = getopt_long(argc, argv, "h:l:p:", long_options, NULL);
    if (s == -1) break;

    switch (s) {
      case 'h':	host = optarg; break;
      case 'l': localhost = optarg; break;
      case 'p': port = atoi(optarg); break;

      case '?':
      default:
	printf(usage, argv[0]);
	exit(2);
    }
  }
  if (optind < argc) {
    printf(usage, argv[0]);
    exit(2);
  }

  /* setup sockets */
  s = WSAStartup(MAKEWORD(2, 2), &supported);
  if (s) {
    fprintf(stderr, "could not initialize sockets\n");
    return 2;
  }
  if (LOBYTE(supported.wVersion) != 2 || HIBYTE(supported.wVersion) != 2) {
    fprintf(stderr, "could not find a useable version of winsock.dll\n");
    WSACleanup();
    return 2;
  }

  /* connect to EVaRT */
  s = evas_evartinit(evas_datahandler);
  if (s) return s;

  s = evas_evartconnect(localhost, host);
  if (s) return s;

  /* setup server */
  servfd = evas_servercreate(port);
  if (servfd == INVALID_SOCKET) goto done;

  s = evas_serverlisten(servfd);

done:
    s = EVaRT2_Exit();
    printf("press any key to terminate...");
    s = getch();

    WSACleanup();
    return 0;
}


/* --- evas_servercreate --------------------------------------------- */

SOCKET
evas_servercreate(unsigned short port)
{
  SOCKET sockfd;
  struct sockaddr_in addr;
  int reuse;

  sockfd = socket(AF_INET, SOCK_DGRAM, 0);
  if (sockfd == INVALID_SOCKET) return INVALID_SOCKET;

  /* set reuse addr option */
  reuse = 1;
  if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR,
		 (char *)&reuse, sizeof(reuse))) {
    perror("cannot set reuseaddr on server socket");
    goto error;
  }

  /* bind */
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = htonl(INADDR_ANY);
  addr.sin_port = htons(port);
  if (bind(sockfd, (struct sockaddr *)&addr, sizeof(struct sockaddr_in)))
    goto error;

  return sockfd;

error:
  fprintf(stderr, "cannot create socket server\n");
  closesocket(sockfd);
  return INVALID_SOCKET;
}


/* --- evas_serverlisten --------------------------------------------- */

static int
evas_serverlisten(SOCKET sockfd)
{
  char buffer[65536];
  evas_msg_t *msg = (evas_msg_t *)buffer;
  struct sockaddr from;
  int fromlen = sizeof(from);
  void *response;
  int bytes;
  int i, s;

  while(1) {
    s = recvfrom(sockfd, buffer, sizeof(buffer), 0, &from, &fromlen);
    if (s <= 0) continue;

    switch(msg->type) {
      case EVAS_NONE: break;

      case EVAS_MSG:
	printf("* %s\n", msg->cmd);
	if (!strncmp(msg->cmd, "hello", 5)) {
	  msg->type = EVAS_MSG;
	  s = snprintf(msg->cmd, sizeof(msg->cmd), PACKAGE_STRING);
	  msg->len = (char *)&msg->cmd[s+1] - (char *)msg;
	  s = sendto(sockfd, buffer, msg->len, 0, &from, fromlen);
	}
	break;

      case EVAS_ACQUIRE:
	printf("* acquire %s\n", msg->cmd);
	if (msg->cmd[0] == '+') {
	  s = EVaRT2_Request("LiveMode", &response, &bytes);
	  s = (s != RC_Okay);
	} else {
	  s = EVaRT2_Request("PauseMode", &response, &bytes);
	  s = (s != RC_Okay);
	}
	msg->type = EVAS_ACQUIRE;
	s = snprintf(msg->cmd, sizeof(msg->cmd), "%d", s);
	msg->len = (char *)&msg->cmd[s+1] - (char *)msg;
	s = sendto(sockfd, buffer, msg->len, 0, &from, fromlen);
	break;

      case EVAS_RECORD:
	printf("* record '%s'\n", msg->cmd);
	if (msg->cmd[0] == '\0') {
	  s = EVaRT2_Request("StopRecording", &response, &bytes);
	  s = (s != RC_Okay);
	} else {
            char rqst[1024];

            snprintf(rqst, sizeof(rqst), "SetOutputName=%s", msg->cmd);
            s = EVaRT2_Request(rqst, &response, &bytes);
	    s = (s != RC_Okay);
	    if (!s) {
	      s = EVaRT2_Request("StartRecording", &response, &bytes);
	      s = (s != RC_Okay);
	    }
	}
	msg->type = EVAS_RECORD;
	s = snprintf(msg->cmd, sizeof(msg->cmd), "%d", s);
	msg->len = (char *)&msg->cmd[s+1] - (char *)msg;
	s = sendto(sockfd, buffer, msg->len, 0, &from, fromlen);
	break;

      case EVAS_FPS:
	printf("* fps\n");
	s = EVaRT2_Request("GetContextFrameRate", &response, &bytes);
	s = (s != RC_Okay);

	msg->type = EVAS_FPS;
	s = snprintf(
	  msg->cmd, sizeof(msg->cmd), "%c%f", s?'-':'+', *(float *)response);
	msg->len = (char *)&msg->cmd[s+1] - (char *)msg;
	s = sendto(sockfd, buffer, msg->len, 0, &from, fromlen);
	break;

      case EVAS_BODY_LIST: {
	sBodyDefs *bd;

	bd = EVaRT2_GetBodyDefs();
	if (!bd) break;
	printf("* listing %d bodies\n", bd->nBodyDefs);

	msg->len =
	  (char *)&msg->body_list.bodies[bd->nBodyDefs] - (char *)msg;
	if (msg->len > sizeof(buffer)) {
	  fprintf(stderr, "* too many bodies (%d)\n", bd->nBodyDefs);
	  break;
	}
	msg->type = EVAS_BODY_LIST;
	msg->body_list.nbodies = bd->nBodyDefs;
	for(i=0; i<bd->nBodyDefs; i++) {
	  strncpy(msg->body_list.bodies[i], bd->BodyDefs[i].szName,
		  sizeof(msg->body_list.bodies[i]));
	}
	sendto(sockfd, buffer, msg->len, 0, &from, fromlen);

	EVaRT2_FreeBodyDefs(bd);
	break;
      }

      case EVAS_BODY_MARKERS_LIST: {
	int body = atoi(msg->cmd);
        unsigned int n;
	sBodyDefs *bd;

	bd = EVaRT2_GetBodyDefs();
	if (!bd) break;
	if (body < 0 || body >= bd->nBodyDefs) {
	  printf("* no such body %d\n", body);
	  break;
	}
        n = bd->BodyDefs[body].nMarkers;

	printf("* listing %d markers for body %d\n", n, body);

	msg->len = (char *)&msg->body_markers_list.markers[n] - (char *)msg;
	if (msg->len > sizeof(buffer)) {
	  fprintf(stderr, "* too many markers (%d)\n", n);
	  break;
	}
	msg->type = EVAS_BODY_MARKERS_LIST;
	msg->body_markers_list.index = body;
	strncpy(msg->body_markers_list.name, bd->BodyDefs[body].szName,
		sizeof(msg->body_markers_list.name));
	msg->body_markers_list.nmarkers = bd->BodyDefs[body].nMarkers;
	for(i=0; i<bd->BodyDefs[body].nMarkers; i++) {
	  strncpy(msg->body_markers_list.markers[i],
		  bd->BodyDefs[body].szMarkerNames[i],
		  sizeof(msg->body_markers_list.markers[i]));
	}
	sendto(sockfd, buffer, msg->len, 0, &from, fromlen);

	EVaRT2_FreeBodyDefs(bd);
	break;
      }

      case EVAS_BODY_SEGMENTS_LIST: {
	int body = atoi(msg->cmd);
        unsigned int n;
	sBodyDefs *bd;

	bd = EVaRT2_GetBodyDefs();
	if (!bd) break;
	if (body < 0 || body >= bd->nBodyDefs) {
	  printf("* no such body %d\n", body);
	  break;
	}
        n = bd->BodyDefs[body].Hierarchy.nSegments;
	printf("* listing %d segments for body %d\n", n, body);

	msg->len = (char *)&msg->body_segments_list.hier[n] - (char *)msg;
	if (msg->len > sizeof(buffer)) {
	  fprintf(stderr, "* too many segments (%d)\n", n);
	  break;
	}
	msg->type = EVAS_BODY_SEGMENTS_LIST;
	msg->body_segments_list.index = body;
	strncpy(msg->body_segments_list.name, bd->BodyDefs[body].szName,
		sizeof(msg->body_segments_list.name));
	msg->body_segments_list.nsegments = n;
	for(i=0; i<n; i++) {
	  strncpy(msg->body_segments_list.hier[i].name,
		  bd->BodyDefs[body].Hierarchy.szSegmentNames[i],
		  sizeof(msg->body_segments_list.hier[i].name));
          msg->body_segments_list.hier[i].parent =
            bd->BodyDefs[body].Hierarchy.iParents[i];
	}
	sendto(sockfd, buffer, msg->len, 0, &from, fromlen);

	EVaRT2_FreeBodyDefs(bd);
	break;
      }

      case EVAS_BODY_DOFS_LIST: {
	int body = atoi(msg->cmd);
        unsigned int n;
	sBodyDefs *bd;

	bd = EVaRT2_GetBodyDefs();
	if (!bd) break;
	if (body < 0 || body >= bd->nBodyDefs) {
	  printf("* no such body %d\n", body);
	  break;
	}
        n = bd->BodyDefs[body].nDofs;
	printf("* listing %d dofs for body %d\n", n, body);

	msg->len = (char *)&msg->body_dofs_list.dofs[n] - (char *)msg;
	if (msg->len > sizeof(buffer)) {
	  fprintf(stderr, "* too many markers (%d)\n", n);
	  break;
	}
	msg->type = EVAS_BODY_DOFS_LIST;
	msg->body_dofs_list.index = body;
	strncpy(msg->body_dofs_list.name, bd->BodyDefs[body].szName,
		sizeof(msg->body_dofs_list.name));
	msg->body_dofs_list.ndofs = n;
	for(i=0; i<n; i++) {
	  strncpy(msg->body_dofs_list.dofs[i],
		  bd->BodyDefs[body].szDofNames[i],
		  sizeof(msg->body_dofs_list.dofs[i]));
	}
	sendto(sockfd, buffer, msg->len, 0, &from, fromlen);

	EVaRT2_FreeBodyDefs(bd);
	break;
      }

      case EVAS_BODY_MARKERS:
      case EVAS_BODY_SEGMENTS:
      case EVAS_BODY_DOFS:
      case EVAS_UNRECOGNIZED_MARKERS:
	switch(msg->cmd[0]) {
	  case '-':
	    s = evas_requestfree(msg->type, &from, msg->cmd+1);
	    break;

	  case '+':
	    s = evas_requestnew(msg->type, &from, msg->cmd+1);
	    break;

	  default:
	    s = evas_requestnew(msg->type, &from, msg->cmd);
	    break;
	}
	break;

      case EVAS_KEEPALIVE:
        s = evas_requestrefresh(&from);
        break;
    }
  }

  return 0;
}


/* --- evas_datahandler ---------------------------------------------- */

static void
evas_datahandler(sFrameOfData *frame)
{
  char buffer[65536];
  evas_msg_t *msg = (evas_msg_t *)buffer;
  unsigned int n;
  int i, j, k, l, r;

  for(i=0; i<sizeof(requests)/sizeof(requests[0]); i++)
    switch(requests[i].type) {
      case EVAS_BODY_MARKERS:
	j = atoi(requests[i].name);
	if (j<0 || j>=frame->nBodies) break;

        n = frame->BodyData[j].nMarkers;
	msg->len = (char *)&msg->body_markers.markers[n] - (char *)msg;
	if (msg->len > sizeof(buffer)) {
	  fprintf(stderr, "too many markers (%d) for body %d:%s\n",
		  n, j, frame->BodyData[j].szName);
	  continue;
	}
	msg->type = EVAS_BODY_MARKERS;
	msg->body_markers.iFrame = frame->iFrame;
	msg->body_markers.fDelay = frame->fDelay;
	evas_gettimeofacq(frame->fDelay,
			  &msg->body_markers.tv_sec, &msg->body_markers.tv_usec);
	msg->body_markers.index = j;
	strncpy(msg->body_markers.name, frame->BodyData[j].szName,
		sizeof(msg->body_markers.name));
	msg->body_markers.nmarkers = n;
	for(k=0; k<msg->body_markers.nmarkers; k++) {
          for(l=0; l<3; l++)
            msg->body_markers.markers[k][l] = frame->BodyData[j].Markers[k][l];
	}

	sendto(servfd, buffer, msg->len, 0,
	       &requests[i].addr, sizeof(requests[i].addr));
	break;

      case EVAS_BODY_SEGMENTS:
	j = atoi(requests[i].name);
	if (j<0 || j>frame->nBodies) break;

        n = frame->BodyData[j].nSegments;
	msg->len = (char *)&msg->body_segments.segments[n] - (char *)msg;
	if (msg->len > sizeof(buffer)) {
	  fprintf(stderr, "too many segments (%d) for body %d:%s\n",
		  n, j, frame->BodyData[j].szName);
	  continue;
	}
	msg->type = EVAS_BODY_SEGMENTS;
	msg->body_segments.iFrame = frame->iFrame;
	msg->body_segments.fDelay = frame->fDelay;
	evas_gettimeofacq(frame->fDelay,
                          &msg->body_segments.tv_sec,
                          &msg->body_segments.tv_usec);
	msg->body_segments.index = j;
	strncpy(msg->body_segments.name, frame->BodyData[j].szName,
		sizeof(msg->body_segments.name));
	msg->body_segments.nsegments = n;
	for(k=0; k<n; k++) {
          for(l=0; l<3; l++) {
            msg->body_segments.segments[k].pos[l] =
              frame->BodyData[j].Segments[k][l];
            msg->body_segments.segments[k].rot[l] =
              frame->BodyData[j].Segments[k][l+3];
          }
          msg->body_segments.segments[k].length =
            frame->BodyData[j].Segments[k][6];
	}

	sendto(servfd, buffer, msg->len, 0,
	       &requests[i].addr, sizeof(requests[i].addr));
	break;

      case EVAS_BODY_DOFS:
	j = atoi(requests[i].name);
	if (j<0 || j>frame->nBodies) break;

        n = frame->BodyData[j].nDofs;
	msg->len = (char *)&msg->body_dofs.dofs[n] - (char *)msg;
	if (msg->len > sizeof(buffer)) {
	  fprintf(stderr, "too many dofs (%d) for body %d:%s\n",
		  n, j, frame->BodyData[j].szName);
	  continue;
	}
	msg->type = EVAS_BODY_DOFS;
	msg->body_dofs.iFrame = frame->iFrame;
	msg->body_dofs.fDelay = frame->fDelay;
	evas_gettimeofacq(frame->fDelay,
                          &msg->body_dofs.tv_sec,
                          &msg->body_dofs.tv_usec);
	msg->body_dofs.index = j;
	strncpy(msg->body_dofs.name, frame->BodyData[j].szName,
		sizeof(msg->body_dofs.name));
	msg->body_dofs.ndofs = n;
	for(k=0; k<n; k++) {
          msg->body_dofs.dofs[k] = frame->BodyData[j].Dofs[k];
	}

	sendto(servfd, buffer, msg->len, 0,
	       &requests[i].addr, sizeof(requests[i].addr));
	break;

      case EVAS_UNRECOGNIZED_MARKERS:
	msg->len =
	  (char *)&msg->unrecognized_markers.markers[frame->nUnidentifiedMarkers] -
	  (char *)msg;
	if (msg->len > sizeof(buffer)) {
	  fprintf(stderr, "too many unidentified markers (%d)\n",
		  frame->nUnidentifiedMarkers);
	  continue;
	}
	msg->type = EVAS_UNRECOGNIZED_MARKERS;
	msg->unrecognized_markers.iFrame = frame->iFrame;
	msg->unrecognized_markers.fDelay = frame->fDelay;
	evas_gettimeofacq(frame->fDelay,
			  &msg->unrecognized_markers.tv_sec,
			  &msg->unrecognized_markers.tv_usec);
	msg->unrecognized_markers.nmarkers = frame->nUnidentifiedMarkers;
	for(k=0; k<msg->unrecognized_markers.nmarkers; k++) {
	  msg->unrecognized_markers.markers[k][0] = frame->UnidentifiedMarkers[k][0];
	  msg->unrecognized_markers.markers[k][1] = frame->UnidentifiedMarkers[k][1];
	  msg->unrecognized_markers.markers[k][2] = frame->UnidentifiedMarkers[k][2];
	}

	sendto(servfd, buffer, msg->len, 0,
	       &requests[i].addr, sizeof(requests[i].addr));
	break;

      default:
        break;
    }

  /* send keepalive rqsts */
  for(i=0, r=0; i<sizeof(requests)/sizeof(requests[0]); i++) {
    if (requests[i].type == EVAS_NONE) continue;
    r++;

    if (strstr(requests[i].name, "ka")) {
      requests[i].keepalive--;
      if (requests[i].keepalive == 0) {
        /* cancel dead requests */
        requests[i].type = EVAS_NONE;
        printf("* cancelling dead request from %s\n",
               evas_addrtoa(&requests[i].addr));
      } else if (requests[i].keepalive < EVAS_TH_KEEPALIVE &&
                 (requests[i].keepalive+i) % EVAS_F_KEEPALIVE == 0) {
        /* send keepalive */
        printf("* requesting keepalive for %s\n",
               evas_addrtoa(&requests[i].addr));
        msg->type = EVAS_KEEPALIVE;
	msg->len = msg->cmd - (char *)msg;
	sendto(servfd, buffer, msg->len, 0,
	       &requests[i].addr, sizeof(requests[i].addr));
      }
    }
  }

  printf("%10d / %3d bodies / %2d streams  \r",
         frame->iFrame, frame->nBodies, r);
}


/* --- evas_requestnew ----------------------------------------------- */

static int
evas_requestnew(evas_msgtype_t type, struct sockaddr *addr, const char *name)
{
  int i;

  for(i=0; i<sizeof(requests)/sizeof(requests[0]); i++)
    if (requests[i].type == EVAS_NONE) {
      requests[i].addr = *addr;
      if (name)
	snprintf(requests[i].name, sizeof(requests[i].name), "%s", name);
      else
	requests[i].name[0] = '\0';
      requests[i].type = type;
      requests[i].keepalive = EVAS_N_KEEPALIVE;
      switch(type) {
        case EVAS_BODY_MARKERS:
          printf("* streaming markers for '%s' to %s\n",
                 requests[i].name, evas_addrtoa(&requests[i].addr));
          break;

        case EVAS_BODY_SEGMENTS:
          printf("* streaming segments for '%s' to %s\n",
                 requests[i].name, evas_addrtoa(&requests[i].addr));
          break;

        case EVAS_BODY_DOFS:
          printf("* streaming dofs for '%s' to %s\n",
                 requests[i].name, evas_addrtoa(&requests[i].addr));
          break;

        default:
          printf("* streaming unknown for '%s' to %s\n",
                 requests[i].name, evas_addrtoa(&requests[i].addr));
          break;
      }
      return 0;
    }

  return -1;
}


/* --- evas_requestfree ---------------------------------------------- */

static int
evas_requestfree(evas_msgtype_t type, struct sockaddr *addr, const char *name)
{
  int i;

  for(i=0; i<sizeof(requests)/sizeof(requests[0]); i++)
    if (requests[i].type == type) {
      if (memcmp(addr, &requests[i].addr, sizeof(*addr))) continue;
      if (name)
	if (strcmp(name, requests[i].name)) continue;

      requests[i].type = EVAS_NONE;
      switch(type) {
        case EVAS_BODY_MARKERS:
          printf("* stopping markers for '%s' to %s\n",
                 requests[i].name, evas_addrtoa(&requests[i].addr));
          break;

        case EVAS_BODY_SEGMENTS:
          printf("* stopping segments for '%s' to %s\n",
                 requests[i].name, evas_addrtoa(&requests[i].addr));
          break;

        case EVAS_BODY_DOFS:
          printf("* stopping dofs for '%s' to %s\n",
                 requests[i].name, evas_addrtoa(&requests[i].addr));
          break;

        default:
          printf("* stopping unknown for '%s' to %s\n",
                 requests[i].name, evas_addrtoa(&requests[i].addr));
          break;
      }
      return 0;
    }

  return -1;
}


/* --- evas_requestrefresh ------------------------------------------------- */

static int
evas_requestrefresh(struct sockaddr *addr)
{
  int i;

  for(i=0; i<sizeof(requests)/sizeof(requests[0]); i++)
    if (requests[i].type != EVAS_NONE) {
      if (memcmp(addr, &requests[i].addr, sizeof(*addr))) continue;

      requests[i].keepalive = EVAS_N_KEEPALIVE;
    }

  return -1;
}


/* --- evas_gettimeofday --------------------------------------------------- */

static void
evas_gettimeofacq(float delay, uint32_t *tv_sec, uint32_t *tv_usec)
{
  static const unsigned long long epochms = 11644473600000000ULL;
  unsigned long long r;
  FILETIME ft;

  GetSystemTimeAsFileTime(&ft);
  r = ft.dwHighDateTime;
  r <<= 32;
  r |= ft.dwLowDateTime;

  /* converting file time to unix epoch */
  r /= 10;  /*convert into microseconds*/
  r -= epochms;

  /* substract delay */
  r -= (int)(delay * 1e6);
  *tv_sec = (long)(r / 1000000UL);
  *tv_usec = (long)(r % 1000000UL);
}


/* --- evas_addrtoa -------------------------------------------------------- */

static const char *
evas_addrtoa(const struct sockaddr *addr)
{
  static char buffer[1024];

  switch (addr->sa_family) {
    case AF_INET: {
      struct sockaddr_in *in = (struct sockaddr_in *)addr;
      snprintf(buffer, sizeof(buffer),
               "%s:%d", inet_ntoa(in->sin_addr), in->sin_port);
      break;
    }

    default:
      snprintf(buffer, sizeof(buffer), "0x%p\n", addr);
      break;
  }

  return buffer;
}
